from typing import List, Mapping, Optional

from aporia.core.types.field import FieldType
from aporia.pyspark.training.pyspark_field_metadata import PySparkFieldMetadata


def create_fields_metadata(
    field_category: str,
    schema: Mapping[str, FieldType],
    schema_to_columns_mapping: Optional[Mapping[str, str]] = None,
) -> List[PySparkFieldMetadata]:
    """A function that prepate the input for the pyspark traning data aggreator.

    Args:
        field_category: The category we create fields for (e.g raw_inputs predicitons)
        schema: The schema configured for that category
        schema_to_columns_mapping: optional mappig of df columns names ot be used instead of schema names

    Returns:
        a list of PySparkFieldMetadata. which the traning data aggreagtor know how to work with
    """
    result = []
    for field_schema_name, field_value_type in schema.items():
        field_column_name = field_schema_name
        if schema_to_columns_mapping is not None:
            field_column_name = schema_to_columns_mapping.get(field_schema_name, field_schema_name)
        result.append(
            PySparkFieldMetadata(
                name=field_schema_name,
                type=field_value_type,
                group=field_category,
                column=field_column_name,
            )
        )
    return result



import logging
from typing import Any, Dict, List, Mapping, Optional, Union

from aporia.core.errors import AporiaError
from aporia.core.http_client import HttpClient
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldSchema

logger = logging.getLogger(LOGGER_NAME)


async def run_create_model_version_query(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    model_type: str,
    features: Mapping[str, Union[FieldSchema, str]],
    predictions: Mapping[str, Union[FieldSchema, str]],
    raw_inputs: Optional[Mapping[str, Union[FieldSchema, str]]] = None,
    metrics: Optional[Dict[str, str]] = None,
    model_data_type: Optional[str] = None,
    labels: Optional[List[str]] = None,
    feature_importance: Optional[Dict[str, float]] = None,
    mapping: Optional[Dict[str, str]] = None,
):
    """Defines the schema for a specific model version.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Model Version
        model_type: Model type
        features: Feature fields
        predictions: Prediction fields
        raw_inputs: Raw input fields.
        metrics: Prediction metric fields.
        model_data_type: Model data type.
        labels: Labels of multi-label, multiclass or binary model. Deprecated.
        feature_importance: Features' importance.
        mapping: General mapping. See Notes.

    Notes:
        Allowed mapping fields are:
            * batch_id_column_name: The name of the key in the `raw_inputs`
                dict that holds the value of the `batch_id`.
            * relevance_column_name: The name of the key in the `predictions`
                dict that holds the value of the `relevance` score of models of type `ranking`.
            * actual_relevance_column_name: The name of the key in the `actuals`
                dict that holds the value of the `relevance` score of models of type `ranking`.
    """
    await _validate_model_schema(
        http_client=http_client,
        model_schema={"features": features, "predictions": predictions, "raw_inputs": raw_inputs},
    )
    model_version_input = {
        "name": model_version,
        "model_type": model_type,
        "version_schema": {
            "features": features,
            "predictions": predictions,
            "raw_inputs": raw_inputs if raw_inputs is not None else None,
            "metrics": metrics,
            "feature_positions": generate_feature_positions(features),
        },
        "model_data_type": model_data_type,
        "labels": labels,
        "feature_importance": feature_importance,
        "mapping": mapping,
    }

    await http_client.post(
        url=f"/models/{model_id}/versions",
        data=model_version_input,
    )


async def _validate_model_schema(
    http_client: HttpClient,
    model_schema: Mapping[str, Optional[Mapping[str, Union[FieldSchema, str]]]],
):
    is_field_schema_as_dict_supported = await _is_field_schema_supported(http_client=http_client)
    for category_name, category_schema in model_schema.items():
        if category_schema is None:
            continue
        for field_name, field_schema in category_schema.items():
            if not is_field_schema_as_dict_supported and not isinstance(field_schema, str):
                raise AporiaError(
                    f"You backend doesn't support dicts in field schemas "
                    f"problematic field is '{field_name}' in '{category_name}'"
                )


async def _is_field_schema_supported(
    http_client: HttpClient,
) -> bool:
    try:
        result = await http_client.get("capabilities/", retry_on_failure=False)
        if isinstance(result, dict):
            return result["is_field_schema_supported"]
        else:
            raise AporiaError("unprocessable result")
    except Exception as e:  # noqa: F841
        logger.debug(
            "/capabilities call failed,returning false for _is_field_schema_supported",
            exc_info=True,
        )
        return False


def generate_feature_positions(features: Mapping[str, Any]) -> Dict[str, int]:
    """Generate features' position from a features dict.

    Args:
        features: features dict

    Returns:
        Features to positions dictionary.
    """
    return {field_name: index for index, field_name in enumerate(features.keys())}



import logging
from typing import Any, Dict, List, Optional, Union

from aporia.core.http_client import HttpClient
from aporia.core.logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


class FieldTrainingData:
    """Field training data."""

    def __init__(
        self,
        field_name: str,
        key: Optional[str] = None,
        bins: Optional[List[Union[float, int, str, bool]]] = None,
        counts: Optional[List[int]] = None,
        min: Optional[float] = None,
        max: Optional[float] = None,
        sum: Optional[float] = None,
        median: Optional[float] = None,
        average: Optional[float] = None,
        std: Optional[float] = None,
        variance: Optional[float] = None,
        num_samples: Optional[int] = None,
        num_missing_values: Optional[int] = None,
        num_posinf_values: Optional[int] = None,
        num_neginf_values: Optional[int] = None,
        num_unique_values: Optional[int] = None,
        num_zero_values: Optional[int] = None,
        mse: Optional[float] = None,
        rmse: Optional[float] = None,
        mae: Optional[float] = None,
        tn: Optional[int] = None,
        fp: Optional[int] = None,
        fn: Optional[int] = None,
        tp: Optional[int] = None,
        recall_score: Optional[float] = None,
        precision_score: Optional[float] = None,
        accuracy: Optional[float] = None,
        f1_score: Optional[float] = None,
    ):
        """Initializes a FieldTrainingData object.

        Args:
            field_name: Field name
            key: Key for dict fields.
            bins: Histogram bin edges.
            counts: Hitsogram.
            min: Minimum value.
            max: Maximum Value.
            sum: Sum of all values.
            median: Median values.
            average: Average value.
            std: Standard deviation value.
            variance: Variance value.
            num_samples: Number of data samples.
            num_missing_values: Number of missing values.
            num_posinf_values: Number of positive infinite values.
            num_neginf_values: Number of negative infinite values.
            num_unique_values: Number of unique values.
            num_zero_values: Number of zero values.
            mse: MSE for numeric predictions (if labels were reported).
            rmse: RMSE for numeric predictions (if labels were reported).
            mae: MAE for numeric predictions (if labels were reported).
            tn: TN for boolean predictions (if labels were reported).
            fp: FP for boolean predictions (if labels were reported).
            fn: FN for boolean predictions (if labels were reported).
            tp: TP for boolean predictions (if labels were reported).
            recall_score: Recall score for boolean predictions (if labels were reported).
            precision_score: Precision score for boolean predictions (if labels were reported).
            accuracy: Accuracy for boolean predictions (if labels were reported).
            f1_score: F1 score for boolean predictions (if labels were reported).
        """
        self.field_name = field_name
        self.key = key
        self.bins = bins
        self.counts = counts
        self.min = min
        self.max = max
        self.sum = sum
        self.median = median
        self.average = average
        self.std = std
        self.variance = variance
        self.num_samples = num_samples
        self.num_missing_values = num_missing_values
        self.num_posinf_values = num_posinf_values
        self.num_neginf_values = num_neginf_values
        self.num_unique_values = num_unique_values
        self.num_zero_values = num_zero_values
        self.mse = mse
        self.rmse = rmse
        self.mae = mae
        self.tn = tn
        self.fp = fp
        self.fn = fn
        self.tp = tp
        self.recall_score = recall_score
        self.precision_score = precision_score
        self.accuracy = accuracy
        self.f1_score = f1_score

    def serialize(self) -> dict:
        """Serializes the field training data to a dict.

        Returns:
            Serialized training data.
        """
        return {
            "fieldName": self.field_name,
            "key": self.key,
            "bins": self.bins,
            "counts": self.counts,
            "min": self.min,
            "max": self.max,
            "sum": self.sum,
            "median": self.median,
            "average": self.average,
            "std": self.std,
            "variance": self.variance,
            "numSamples": self.num_samples,
            "numMissingValues": self.num_missing_values,
            "numPosinfValues": self.num_posinf_values,
            "numNeginfValues": self.num_neginf_values,
            "numUniqueValues": self.num_unique_values,
            "numZeroValues": self.num_zero_values,
            "mse": self.mse,
            "rmse": self.rmse,
            "mae": self.mae,
            "tn": self.tn,
            "fp": self.fp,
            "fn": self.fn,
            "tp": self.tp,
            "recallScore": self.recall_score,
            "precisionScore": self.precision_score,
            "accuracy": self.accuracy,
            "f1Score": self.f1_score,
        }

    def serialize_backward_compatibility(self) -> dict:
        """Serializes the field training data to a dict without the new performance metric.

        Returns:
            Serialized training data without the new performance metric for backward compatibility.
        """
        return {
            "fieldName": self.field_name,
            "key": self.key,
            "bins": self.bins,
            "counts": self.counts,
            "min": self.min,
            "max": self.max,
            "sum": self.sum,
            "median": self.median,
            "average": self.average,
            "std": self.std,
            "variance": self.variance,
            "numSamples": self.num_samples,
            "numMissingValues": self.num_missing_values,
            "numPosinfValues": self.num_posinf_values,
            "numNeginfValues": self.num_neginf_values,
            "numUniqueValues": self.num_unique_values,
            "numZeroValues": self.num_zero_values,
        }


async def log_training_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[FieldTrainingData],
    predictions: Optional[List[FieldTrainingData]] = None,
    labels: Optional[List[FieldTrainingData]] = None,
    raw_inputs: Optional[List[FieldTrainingData]] = None,
):
    """Reports training data.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Mode version
        features: Training set features.
        predictions: Training set predictions.
        labels: Training set labels.
        raw_inputs: Training set raw inputs.
    """
    query = """
        mutation LogTrainingSet(
            $modelId: String!,
            $modelVersion: String!,
            $features: [FieldTrainingData]!
            $predictions: [FieldTrainingData]
            $labels: [FieldTrainingData]!
            $rawInputs: [FieldTrainingData]
        ) {
            logTrainingSet(
                modelId: $modelId,
                modelVersion: $modelVersion,
                predictions: $predictions
                features: $features
                labels: $labels
                rawInputs: $rawInputs
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "predictions": [field_data.serialize() for field_data in predictions]
        if predictions is not None
        else None,
        "features": [field_data.serialize() for field_data in features],
        "labels": [field_data.serialize() for field_data in labels] if labels is not None else None,
        "rawInputs": [field_data.serialize() for field_data in raw_inputs]
        if raw_inputs is not None
        else None,
    }

    # For on_prem deployment with old controller version that don't support predictions
    backward_compatibility_query = """
        mutation LogTrainingSet(
            $modelId: String!,
            $modelVersion: String!,
            $features: [FieldTrainingData]!
            $labels: [FieldTrainingData]!
            $rawInputs: [FieldTrainingData]
        ) {
            logTrainingSet(
                modelId: $modelId,
                modelVersion: $modelVersion,
                features: $features
                labels: $labels
                rawInputs: $rawInputs
            ) {
                warnings
            }
        }
    """
    backward_compatibility_variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": [field_data.serialize_backward_compatibility() for field_data in features],
        "labels": [field_data.serialize_backward_compatibility() for field_data in labels]
        if labels is not None
        else None,
        "rawInputs": [field_data.serialize_backward_compatibility() for field_data in raw_inputs]
        if raw_inputs is not None
        else None,
    }

    try:
        result = await http_client.graphql(query, variables)
        for warning in result["logTrainingSet"]["warnings"]:
            logger.warning(warning)

    # Handling new sdk with old controller that don't receive predictions in training
    # The compatibility is for versions before inserting ticket APR-1080
    except Exception as exp:
        if "Unknown argument 'predictions' on field 'Mutation.logTrainingSet'" in str(exp):
            logger.warning(
                "Training set's predictions reporting is unsupported. Please update your controller version."
            )
            result = await http_client.graphql(
                backward_compatibility_query, backward_compatibility_variables
            )
            for warning in result["logTrainingSet"]["warnings"]:
                logger.warning(warning)
        elif "Field 'mse' is not defined by type 'FieldTrainingData'" in str(exp):
            # We don't log a warning cause the user didn't try to report predictions
            result = await http_client.graphql(
                backward_compatibility_query, backward_compatibility_variables
            )
            for warning in result["logTrainingSet"]["warnings"]:
                logger.warning(warning)
        else:
            raise exp


async def log_test_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[FieldTrainingData],
    predictions: List[FieldTrainingData],
    labels: List[FieldTrainingData],
    raw_inputs: Optional[List[FieldTrainingData]] = None,
):
    """Reports test data.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Mode version
        features: Test set features.
        predictions: Test set features.
        labels: Test set labels.
        raw_inputs: Test set raw inputs.
    """
    query = """
        mutation LogTestSet(
            $modelId: String!,
            $modelVersion: String!,
            $features: [FieldTrainingData]!
            $predictions: [FieldTrainingData]!
            $labels: [FieldTrainingData]!
            $rawInputs: [FieldTrainingData]
        ) {
            logTestSet(
                modelId: $modelId,
                modelVersion: $modelVersion,
                features: $features
                predictions: $predictions
                labels: $labels
                rawInputs: $rawInputs
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": [field_data.serialize() for field_data in features],
        "predictions": [field_data.serialize() for field_data in predictions],
        "labels": [field_data.serialize() for field_data in labels],
        "rawInputs": [field_data.serialize() for field_data in raw_inputs]
        if raw_inputs is not None
        else None,
    }

    backward_compatibility_variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": [field_data.serialize_backward_compatibility() for field_data in features],
        "predictions": [
            field_data.serialize_backward_compatibility() for field_data in predictions
        ],
        "labels": [field_data.serialize_backward_compatibility() for field_data in labels],
        "rawInputs": [field_data.serialize_backward_compatibility() for field_data in raw_inputs]
        if raw_inputs is not None
        else None,
    }

    try:
        result = await http_client.graphql(query, variables)
        for warning in result["logTestSet"]["warnings"]:
            logger.warning(warning)
    # Handling new sdk with old controller that don't receive predictions in training
    # The compatibility is for versions before inserting ticket APR-1080
    except Exception as exp:
        if "Field 'mse' is not defined by type 'FieldTrainingData'" in str(exp):
            logger.warning(
                "Test set's performance metrics are unsupported. Please update your controller version."
            )
            result = await http_client.graphql(query, backward_compatibility_variables)
            for warning in result["logTestSet"]["warnings"]:
                logger.warning(warning)
        else:
            raise exp


async def log_training_sample_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[Dict[str, Any]],
    labels: Optional[List[Any]] = None,
    raw_inputs: Optional[List[Any]] = None,
):
    """Reports training sample.

    Args:
        http_client: Http client
        model_id: Model ID.
        model_version: Mode version.
        features: List of features.
        labels: List of labels.
        raw_inputs: List of raw inputs.
    """
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/training_sample",
        data={"features": features, "labels": labels, "raw_inputs": raw_inputs},
    )



from typing import Any, Mapping, Optional, Union

from aporia.core.errors import AporiaError
from aporia.inference.inference_model import InferenceModel

try:
    import numpy
    import pandas

    import aporia.training
    from aporia.training.training_model import SAMPLE_SIZE

    ndarray = numpy.ndarray
    DataFrame = pandas.DataFrame
    TRAINING_ENABLED = True

except ImportError:
    ndarray = Any  # type: ignore
    DataFrame = Any
    TRAINING_ENABLED = False
    SAMPLE_SIZE = 1000

try:
    import numpy
    import pandas
    import pyspark.sql

    import aporia.pyspark

    PySparkDataFrame = pyspark.sql.DataFrame
    PYSPARK_ENABLED = True

except ImportError:
    PySparkDataFrame = Any  # type: ignore
    PYSPARK_ENABLED = False


class Model(InferenceModel):
    """Model object for logging model events."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes a model object.

        Args:
            model_id: Model identifier, as received from the Aporia dashboard.
            model_version: Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        super().__init__(model_id=model_id, model_version=model_version)

        self._training = None
        if TRAINING_ENABLED:
            self._training = aporia.training.TrainingModel(
                model_id=model_id, model_version=model_version
            )

        self._pyspark_inference = None
        if PYSPARK_ENABLED:
            self._pyspark_inference = aporia.pyspark.PySparkInferenceModel(inference_model=self)

        self._pyspark_training = None
        if PYSPARK_ENABLED and self._training is not None:
            self._pyspark_training = aporia.pyspark.PySparkTrainingModel(
                training_model=self._training
            )

    def __getstate__(self) -> dict:
        """Returns picklable representation of the Model object."""
        return {
            "model_id": self.model_id,
            "model_version": self.model_version,
        }

    def __setstate__(self, state: dict):
        """Loads Model object from pickled state.

        Args:
            state: Pickled state
        """
        self.__init__(  # type: ignore
            model_id=state["model_id"],
            model_version=state["model_version"],
        )

    def log_training_set(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        predictions: Optional[DataFrame] = None,  # type: ignore
        labels: Optional[DataFrame] = None,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
        log_sample: bool = True,
        sample_size: int = SAMPLE_SIZE,
    ):
        """Logs aggregations of the whole set and logs a sample of the data.

        Args:
            features: Training set features
            predictions: Training set predictions
            labels: Training set labels
            raw_inputs: Training set raw inputs.
            log_sample: Whether to log a sample of the data.
            sample_size: Number of records to sample.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * predictions -> predictions
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging training data failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log training data. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_training_set(
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
                log_sample=log_sample,
                sample_size=sample_size,
            )

    def log_training_sample(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        labels: DataFrame,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
        sample_size: int = SAMPLE_SIZE,
    ):
        """Logs a sample of the training data.

        Args:
            features: Training set features
            labels: Training set labels
            raw_inputs: Training set raw_inputs
            sample_size: Number of records to sample

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * labels -> predictions
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging training samples failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log training samples. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_training_data(
                features=features, labels=labels, raw_inputs=raw_inputs, sample_size=sample_size
            )

    def log_test_set(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        predictions: DataFrame,  # type: ignore
        labels: DataFrame,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
        confidences: Optional[ndarray] = None,  # type: ignore
    ):
        """Logs test data.

        Args:
            features: Test set features
            predictions: Test set predictions
            labels: Test set labels
            raw_inputs: Test set raw inputs.
            confidences: Confidence values for the test predictions.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * predictions -> predictions
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging test data failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log test data. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_test_set(
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
                confidences=confidences,
            )

    def log_batch_pyspark_raw_inputs(
        self,
        ids: PySparkDataFrame,  # type: ignore
        raw_inputs: PySparkDataFrame,  # type: ignore
    ):
        """Logs raw inputs of multiple predictions.

        Args:
            ids: Prediction identifiers
            raw_inputs: Raw inputs of each prediction

        Notes:
            * The ids dataframe must contain exactly one column
            * The ids and raw_inputs dataframes must have the same number of rows
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark raw inputs failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log raw inputs from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )

            self._pyspark_inference.log_batch_pyspark_raw_inputs(ids=ids, raw_inputs=raw_inputs)

    def log_batch_pyspark_actuals(
        self,
        ids: PySparkDataFrame,  # type: ignore
        actuals: PySparkDataFrame,  # type: ignore
    ):
        """Logs actual values of multiple predictions.

        Args:
            ids: Prediction identifiers
            actuals: Actual prediction results of each prediction

        Notes:
            * The ids dataframe must contain exactly one column
            * The ids and actuals dataframes must have the same number of rows
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark actuals failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log actuals from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )

            self._pyspark_inference.log_batch_pyspark_actuals(ids=ids, actuals=actuals)

    def log_batch_pyspark_prediction(
        self,
        data: PySparkDataFrame,
        id_column: Optional[str] = None,
        timestamp_column: Optional[str] = None,
        features: Optional[Mapping[str, str]] = None,
        predictions: Optional[Mapping[str, str]] = None,
        raw_inputs: Optional[Mapping[str, str]] = None,
        labels: Optional[Mapping[str, str]] = None,
        spark_options: Optional[Mapping[str, str]] = None,
    ):
        """Logs multiple predictions.

        Args:
            data: PsSpark dataframe
            id_column: Optional dataframe column to use as an id.
            timestamp_column: Optional dataframe column to use as an id.
            features: A mapping of feature names (from the schema) to dataframe columns
            predictions: A mapping of predictions names (from the schema) to dataframe columns
            raw_inputs: A mapping of raw input names (from the schema) to dataframe columns
            labels: A mapping of label names (from the schema) to dataframe columns
            spark_options: Optional configuration extension for spark elastic connector. See
                https://www.elastic.co/guide/en/elasticsearch/hadoop/master/configuration.html
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark predictions failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log predictions from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )
            from aporia.experimental.model_versions_api import get_model_versions

            version_id = next(
                version.id
                for version in get_model_versions(self.model_id)
                if version.name == self.model_version
            )
            self._pyspark_inference.log_batch_pyspark_prediction(
                version_id=version_id,
                model=self,
                data=data,
                id_column=id_column,
                features=features,
                predictions=predictions,
                timestamp_column=timestamp_column,
                raw_inputs=raw_inputs,
                labels=labels,
                spark_options=spark_options,
            )

    def log_streaming_pyspark_prediction(
        self,
        data: PySparkDataFrame,
        id_column: Optional[str] = None,
        timestamp_column: Optional[str] = None,
        features: Optional[Mapping[str, str]] = None,
        predictions: Optional[Mapping[str, str]] = None,
        raw_inputs: Optional[Mapping[str, str]] = None,
        labels: Optional[Mapping[str, str]] = None,
        spark_options: Optional[Mapping[str, str]] = None,
    ):
        """Logs stream of predictions.

        Args:
            data: PsSpark dataframe
            id_column: Optional dataframe column to use as an id.
            timestamp_column: Optional dataframe column to use as an id.
            features: A mapping of feature names (from the schema) to dataframe columns
            predictions: A mapping of predictions names (from the schema) to dataframe columns
            raw_inputs: A mapping of raw input names (from the schema) to dataframe columns
            labels: A mapping of label names (from the schema) to dataframe columns
            spark_options: Optional configuration extension for spark elastic connector. See
                https://www.elastic.co/guide/en/elasticsearch/hadoop/master/configuration.html
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark predictions failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log predictions from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )
            from aporia.experimental.model_versions_api import get_model_versions

            version_id = next(
                version.id
                for version in get_model_versions(self.model_id)
                if version.name == self.model_version
            )

            return self._pyspark_inference.log_streaming_pyspark_prediction(
                version_id=version_id,
                model=self,
                data=data,
                id_column=id_column,
                features=features,
                predictions=predictions,
                timestamp_column=timestamp_column,
                raw_inputs=raw_inputs,
                labels=labels,
                spark_options=spark_options,
            )

    def log_pyspark_training_set(
        self,
        data: PySparkDataFrame,
        features: Optional[Mapping[str, str]] = None,
        predictions: Optional[Mapping[str, str]] = None,
        raw_inputs: Optional[Mapping[str, str]] = None,
    ):
        """Logs training data from PySpark DataFrames.

        Args:
            data: PySparkDataFrame all the data
            features: Optional[Mapping[str,str]]
                mapping of column to features as configured in the schema.
                Key is the name in the schema and the value is the name of the corresponding dataframe column
            predictions:Optional[Mapping[str,str]]
                mapping of column to predictions as configured in the schema.
                Key is the name in the schema and the value is the name of the corresponding dataframe column
            raw_inputs: Optional[Mapping[str, str]]
                mapping of column to raw inputs as configured in the schema.
                Key is the name in the schema and the value is the name of the corresponding dataframe column

        """
        with self.handle_error("{}"):
            if self._pyspark_training is None:
                raise AporiaError(
                    short_message="Logging training data failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log training data from "
                        "pyspark dataframes. Install it with `pip install aporia[pyspark]` "
                        "and try again."
                    ),
                )

            self._pyspark_training.log_pyspark_training_set(
                data=data,
                features_schema_df_mapping=features,
                predictions_schema_df_mapping=predictions,
                raw_inputs_schema_df_mapping=raw_inputs,
            )

    def log_pyspark_test_set(
        self,
        features: PySparkDataFrame,  # type: ignore
        predictions: PySparkDataFrame,  # type: ignore
        labels: PySparkDataFrame,  # type: ignore
        raw_inputs: Optional[PySparkDataFrame] = None,  # type: ignore
    ):
        """Logs test data from PySpark DataFrames.

        Args:
            features: Test set features
            predictions: Test set predictions
            labels: Test set labels
            raw_inputs: Test set raw inputs.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * predictions -> predictions
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._pyspark_training is None:
                raise AporiaError(
                    short_message="Logging test data failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log test data from "
                        "pyspark dataframes. Install it with `pip install aporia[pyspark]` "
                        "and try again."
                    ),
                )

            self._pyspark_training.log_pyspark_test_set(
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
            )



from typing import Any, Dict, Mapping, Optional

from pyspark.sql import Column, DataFrame
import pyspark.sql.functions as F

from aporia.core.context import get_context
from aporia.core.types.field import FieldCategory


def timestamp_to_iso8601(timestamp: Column) -> Column:
    """Converts a timestamp to ISO-8601 format."""
    value = F.date_format(timestamp, "yyyy-MM-dd'T'HH:mm:ss.SSSZ")
    return F.regexp_replace(value, r"\+0000", "+00:00")


def apply_dataset_schema(
    data: DataFrame,
    features: Dict[str, str],
    predictions: Dict[str, str],
    raw_inputs: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    id_column: Optional[str] = None,
    timestamp_column: Optional[str] = None,
) -> DataFrame:
    """Renames and groups dataframe columns according to a schema.

    Args:
        data: Dataframe
        features: Feature name->column mapping
        predictions: Prediction name->column mapping
        raw_inputs: Raw inputs name->column mapping
        labels: Label name->column mapping
        id_column: Name of ID column
        timestamp_column: Name of timestamp column

    Returns:
        Modified dataframe
    """
    select_params = []

    if id_column is not None:
        select_params.append(F.col(id_column).alias("id"))
    else:
        select_params.append(F.expr("uuid()").alias("id"))

    if timestamp_column is not None:
        select_params.append(F.col(timestamp_column).alias("timestamp"))
    else:
        select_params.append(F.current_timestamp().alias("timestamp"))

    feature_renames = [F.col(column).alias(name) for name, column in features.items()]
    select_params.append(F.struct(*feature_renames).alias("features"))

    prediction_renames = [F.col(column).alias(name) for name, column in predictions.items()]
    select_params.append(F.struct(*prediction_renames).alias("predictions"))

    if raw_inputs is not None:
        raw_input_renames = [F.col(column).alias(name) for name, column in raw_inputs.items()]
        select_params.append(F.struct(*raw_input_renames).alias("raw_inputs"))

    if labels is not None:
        label_renames = [F.col(column).alias(name) for name, column in labels.items()]
        select_params.append(F.struct(*label_renames).alias("actuals"))

    return data.select(*select_params)


def add_metadata_columns(
    data: DataFrame,
    model_id: str,
    version_name: str,
    version_id: str,
    environment: str,
    environment_id: str,
    add_occurred_at: bool,
) -> DataFrame:
    """Adds metadata columns to a dataframe.

    Args:
        data: Dataframe
        model_id: Model ID
        version_name: Model version name
        version_id: Model version ID
        environment: Environment name
        environment_id: Environment ID
        add_occurred_at: True to add an `occurred_at` column, False otherwise

    Returns:
        Modified dataframe
    """
    current_timestamp = timestamp_to_iso8601(F.current_timestamp())

    data = data.select(
        "*",
        F.lit(model_id).alias("model_id"),
        F.lit(version_name).alias("model_version"),
        F.lit(version_id).alias("model_version_id"),
        F.lit(environment).alias("environment"),
        F.lit(environment_id).alias("environment_id"),
        current_timestamp.alias("updated_at"),
        current_timestamp.alias("created_at"),
        F.date_format(data["timestamp"], "yyyy_MM").alias("date"),
    )

    if add_occurred_at:
        data = data.select("*", timestamp_to_iso8601(data["timestamp"]).alias("occurred_at"))

    return data


def _internal_log_prediction(
    apr_model: Any,
    version_id: str,
    data: DataFrame,
    timestamp_column: Optional[str] = None,
    id_column: Optional[str] = None,
    features: Optional[Mapping[str, str]] = None,
    predictions: Optional[Mapping[str, str]] = None,
    raw_inputs: Optional[Mapping[str, str]] = None,
    labels: Optional[Mapping[str, str]] = None,
    spark_options: Optional[Mapping[str, str]] = None,
):
    """Internal implementation of pyspark prediction logging.

    Args:
        apr_model: Aporia Model object
        version_id: Model version ID
        data: Dataframe containing prediction data
        timestamp_column: Name of timestamp column in data
        id_column: Name of id column in data
        features: Mapping of feature name->column
        predictions: Mapping of prediction name->column
        raw_inputs: Mapping of raw_input name->column
        labels: Mapping of label name->column
        spark_options: Additional spark configuration options
    """
    # Fetch cluster and environment ID
    context = get_context()

    response = context.event_loop.run_coroutine(
        context.http_client.post(
            url="environments",
            data={"name": apr_model._config.environment},
        )
    )

    cluster_id = response["cluster_id"]
    environment_id = response["environment_id"]

    # Fetch model version schema
    version_schema = apr_model._training.get_version_schema()

    # Apply default schema + overrides
    features = {
        **{field: field for field in version_schema[FieldCategory.FEATURES].keys()},
        **(features if features is not None else {}),
    }

    predictions = {
        **{field: field for field in version_schema[FieldCategory.PREDICTIONS].keys()},
        **(predictions if predictions is not None else {}),
    }

    if FieldCategory.RAW_INPUTS in version_schema:
        raw_inputs = {
            **{field: field for field in version_schema.get(FieldCategory.RAW_INPUTS, {}).keys()},
            **(raw_inputs if raw_inputs is not None else {}),
        }

    # Add types (e.g my_field -> numeric_my_field)
    features = {
        f"{version_schema[FieldCategory.FEATURES][key].value}_{key}": value
        for key, value in features.items()
    }
    predictions = {
        f"{version_schema[FieldCategory.PREDICTIONS][key].value}_{key}": value
        for key, value in predictions.items()
    }

    if labels is not None:
        labels = {
            f"{version_schema[FieldCategory.PREDICTIONS][key].value}_{key}": value
            for key, value in labels.items()
        }

    if raw_inputs is not None:
        raw_inputs = {
            f"{version_schema[FieldCategory.RAW_INPUTS][key].value}_{key}": value
            for key, value in raw_inputs.items()
        }

    data = apply_dataset_schema(
        data=data,
        id_column=id_column,
        timestamp_column=timestamp_column,
        features=features,
        predictions=predictions,
        raw_inputs=raw_inputs,
        labels=labels,
    )

    data = add_metadata_columns(
        data=data,
        model_id=apr_model.model_id,
        version_name=apr_model.model_version,
        version_id=version_id,
        environment=apr_model._config.environment,
        environment_id=environment_id,
        add_occurred_at=(predictions is not None),
    )

    base_index = f"aporia_{cluster_id}_pred_{apr_model.model_id}_{version_id}_{environment_id}"

    options = {
        "es.nodes.wan.only": "true",
        "es.port": str(apr_model._config.port),
        "es.nodes": str(apr_model._config.host),
        "es.net.http.header.Authorization": f"Bearer {apr_model._config.token}",
        "es.nodes.path.prefix": "/v1/es/",
        "es.resource": f"{base_index}_{{date}}",
        "es.write.operation": "index"
        if features is not None and predictions is not None
        else "upsert",
        "es.mapping.id": "id",
        "es.mapping.exclude": "id,timestamp,date",
        "es.batch.write.refresh": "false",
        "es.net.ssl": "true" if apr_model._config.port == 443 else "false",
        "es.net.ssl.cert.allow.self.signed": "false" if apr_model._config.verify_ssl else "true",
        **(spark_options if spark_options is not None else {}),
    }

    return data, options


def log_batch_prediction(
    apr_model: Any,
    version_id: str,
    data: DataFrame,
    timestamp_column: Optional[str] = None,
    id_column: Optional[str] = None,
    features: Optional[Mapping[str, str]] = None,
    predictions: Optional[Mapping[str, str]] = None,
    raw_inputs: Optional[Mapping[str, str]] = None,
    labels: Optional[Mapping[str, str]] = None,
    spark_options: Optional[Mapping[str, str]] = None,
):
    """Logs a batch of predictions using pyspark.

    Args:
        apr_model: Aporia Model object
        version_id: Model version ID
        data: Dataframe containing prediction data
        timestamp_column: Name of timestamp column in data
        id_column: Name of id column in data
        features: Mapping of feature name->column
        predictions: Mapping of prediction name->column
        raw_inputs: Mapping of raw_input name->column
        labels: Mapping of label name->column
        spark_options: Additional spark configuration options
    """

    data, options = _internal_log_prediction(
        apr_model=apr_model,
        version_id=version_id,
        data=data,
        timestamp_column=timestamp_column,
        id_column=id_column,
        features=features,
        predictions=predictions,
        raw_inputs=raw_inputs,
        labels=labels,
        spark_options=spark_options,
    )

    data.write.format("org.elasticsearch.spark.sql").options(**options).mode("append").save()


def log_streaming_prediction(
    apr_model: Any,
    version_id: str,
    data: DataFrame,
    timestamp_column: Optional[str] = None,
    id_column: Optional[str] = None,
    features: Optional[Mapping[str, str]] = None,
    predictions: Optional[Mapping[str, str]] = None,
    raw_inputs: Optional[Mapping[str, str]] = None,
    labels: Optional[Mapping[str, str]] = None,
    spark_options: Optional[Mapping[str, str]] = None,
):
    """Logs a stream of predictions using pyspark.

    Args:
        apr_model: Aporia Model object
        version_id: Model version ID
        data: Dataframe containing prediction data
        timestamp_column: Name of timestamp column in data
        id_column: Name of id column in data
        features: Mapping of feature name->column
        predictions: Mapping of prediction name->column
        raw_inputs: Mapping of raw_input name->column
        labels: Mapping of label name->column
        spark_options: Additional spark configuration options
    """

    data, options = _internal_log_prediction(
        apr_model=apr_model,
        version_id=version_id,
        data=data,
        timestamp_column=timestamp_column,
        id_column=id_column,
        features=features,
        predictions=predictions,
        raw_inputs=raw_inputs,
        labels=labels,
        spark_options=spark_options,
    )

    return data.writeStream.format("es").options(**options)



import logging
from typing import Any, Dict, List, Sequence, Tuple

from pyspark.sql import Column, DataFrame
import pyspark.sql.functions as F
from pyspark.sql.types import DoubleType, IntegerType

from aporia.core.errors import AporiaError
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldType
from aporia.pyspark.training.pyspark_aggregation_alias import (
    create_alias,
    parse_aggregation_results_to_field_training_data_by_group,
)
from aporia.pyspark.training.pyspark_field_metadata import PySparkFieldMetadata
from aporia.pyspark.training.utils import expand_dicts_keys_attributes_to_regular_fields
from aporia.training.api.log_training import FieldTrainingData

logger = logging.getLogger(LOGGER_NAME)


def get_positive_infinity() -> Column:
    """Returns pyspark positive infinity."""
    return F.lit("+Infinity").cast("double")


def get_negative_infinity() -> Column:
    """Returns pyspark negative infinity."""
    return F.lit("-Infinity").cast("double")


def percentile_approximate_accuracy() -> Column:
    """Returns pyspark proximity."""
    return F.lit(100000)


TRAINING_BIN_COUNT = 50
HISTOGRAMABLE_TYPES = [
    FieldType.NUMERIC,
    FieldType.BOOLEAN,
    FieldType.STRING,
    FieldType.CATEGORICAL,
]

NOT_A_DICT_KEY = "NOT_A_DICT_KEY"


class _TrainingAggregatorPySpark:
    def __init__(self, data: DataFrame, fields: List[PySparkFieldMetadata]):
        """Initializes the training aggregator.

        Args:
            data: Data to compute training aggregations on
            fields: Field metadata
        """
        # Cache the dataframe to improve performance
        self.initial_data_cached = data.cache()
        # Enrich the dataframe by expanding dict fields
        (normalized_data, normalized_fields) = self._normalize_dict_fields(df=data, fields=fields)
        self.data = normalized_data
        self.fields = normalized_fields
        # Cache the modified dataframe to improve performance
        self.data.cache()

    def __del__(self):
        """Clean the cached dataframe from the init step."""
        self.initial_data_cached.unpersist()
        self.data.unpersist()

    def aggregate_training_fields(self) -> Dict[str, List[FieldTrainingData]]:
        """This function gets list of fields and calculate the training aggregations from them."""
        aggregation_columns = self._generate_all_aggregations(self.fields)
        aggregations_by_alias = dict()
        for agg_col in aggregation_columns:
            logger.info("aggregating col:")
            logger.info(agg_col)
            agg = self.data.select(agg_col).collect()[0].asDict()
            aggregations_by_alias.update(agg)
        logger.info("done with aggregation that are not histogram")

        aggregations_by_alias.update(self._compute_group_histograms_array(self.data, self.fields))

        fields_training_data_by_group = parse_aggregation_results_to_field_training_data_by_group(
            aggregations_by_alias
        )

        return fields_training_data_by_group

    def _normalize_dict_fields(
        self, df: DataFrame, fields: List[PySparkFieldMetadata]
    ) -> Tuple[DataFrame, List[PySparkFieldMetadata]]:
        dict_type_fields = [field for field in fields if field.type is FieldType.DICT]

        field_metadata_and_column_pairs = expand_dicts_keys_attributes_to_regular_fields(
            df, dict_type_fields
        )

        # add the keys of the dict field to the dataframe
        dict_keys_cols = [col for _, col in field_metadata_and_column_pairs]
        all_cols = [F.col(col) for col in df.columns] + dict_keys_cols
        # add the fields metadata for each dict field.
        fields_with_expanded_dicts = [
            *[field for field in fields if field.type is not FieldType.DICT],
            *[field_metadata for field_metadata, _ in field_metadata_and_column_pairs],
        ]

        return (df.select(all_cols), fields_with_expanded_dicts)

    def _get_column_valid_values(self, column: Column) -> Column:
        return F.when(
            condition=(self._get_numeric_valid_condition(column)),
            value=column,
        )

    def _get_numeric_valid_condition(self, column: Column) -> Column:
        return (
            ~F.isnan(column)
            & ~column.isNull()
            & ~column.isin([get_positive_infinity(), get_negative_infinity()])
        )

    def _calculate_histogram(
        self, data: DataFrame, field: PySparkFieldMetadata
    ) -> Tuple[Sequence[Any], List[int]]:
        logger.debug("Generating histogram aggregations for:" f"{field}")

        if field.type is FieldType.NUMERIC:
            bins, counts = (
                data.select(field.column)
                .replace([float("inf"), float("-inf")], None)
                .rdd.flatMap(lambda x: x)
                .histogram(TRAINING_BIN_COUNT)
            )

        elif field.type in [FieldType.BOOLEAN, FieldType.STRING, FieldType.CATEGORICAL]:
            distinct_values = (
                data.select(field.column)
                .replace([float("inf"), float("-inf")], None)
                .dropna()
                .groupBy(field.column)
                .count()
                .orderBy(F.col("count").desc())
                .orderBy(field.column)
            )

            bins = []
            counts = []
            for row in distinct_values.collect():
                bins.append(row[field.column])
                counts.append(row["count"])

        else:
            raise AporiaError(f"Unsupported data type {field.type} of column {field.column}")

        return bins, counts

    def _compute_group_histograms_array(
        self, data: DataFrame, fields: List[PySparkFieldMetadata]
    ) -> dict:
        result: Dict[str, Any] = {}
        for field in fields:
            if field.type in HISTOGRAMABLE_TYPES:
                bins, counts = self._calculate_histogram(data=data, field=field)
                result[create_alias("bins", field)] = bins
                result[create_alias("counts", field)] = counts
                if field.type is FieldType.BOOLEAN:
                    result[create_alias("num_unique_values", field)] = len(bins)

        return result

    def _generate_all_aggregations(self, fields: List[PySparkFieldMetadata]) -> List[Column]:
        result: List[Column] = []
        for field in fields:
            result.extend(self._generate_column_aggregations(field))
        return result

    def _generate_column_aggregations(self, field: PySparkFieldMetadata) -> List[Column]:
        logger.debug("Generating aggregations for:" f"{field}")
        if field.type in [FieldType.DATETIME, FieldType.VECTOR, FieldType.TEXT, FieldType.TENSOR]:
            return self._aggregate_any_column(field)
        elif field.type == FieldType.NUMERIC:
            return self._aggregate_numeric_column(field)
        elif field.type == FieldType.BOOLEAN:
            return self._aggregate_boolean_column(field)
        elif field.type in [FieldType.STRING, FieldType.CATEGORICAL]:
            return self._aggregate_categorical_column(field)

        raise AporiaError(f"Unsupported field type {field.type} of column {field.column}")

    def _aggregate_numeric_column(self, field: PySparkFieldMetadata) -> List[Column]:
        column = field.column
        valid_values = self._get_column_valid_values(F.col(column))

        return [
            F.max(valid_values).alias(create_alias("max", field)),
            F.min(valid_values).alias(create_alias("min", field)),
            F.sum(valid_values).alias(create_alias("sum", field)),
            F.avg(valid_values).alias(create_alias("average", field)),
            F.stddev(valid_values).alias(create_alias("std", field)),
            F.variance(valid_values).alias(create_alias("variance", field)),
            F.count(valid_values).alias(create_alias("num_samples", field)),
            F.count(F.when(F.col(column) == 0, column)).alias(
                create_alias("num_zero_values", field)
            ),
            F.percentile_approx(valid_values, 0.5, percentile_approximate_accuracy()).alias(
                create_alias("median", field)
            ),
            F.count(F.when(F.isnan(column) | F.col(column).isNull(), column)).alias(
                create_alias("num_missing_values", field)
            ),
            F.count(F.when(F.col(column) == (get_positive_infinity()), column)).alias(
                create_alias("num_posinf_values", field)
            ),
            F.count(F.when(F.col(column) == get_negative_infinity(), column)).alias(
                create_alias("num_neginf_values", field)
            ),
        ]

    def _aggregate_boolean_column(self, field: PySparkFieldMetadata) -> List[Column]:
        all_values = F.col(field.column)
        colType = self.data.schema[field.column].dataType
        isNumricCol = isinstance(colType, DoubleType) or isinstance(colType, IntegerType)

        valid_condition = ~all_values.isNull()
        if isNumricCol:
            valid_condition = self._get_numeric_valid_condition(all_values)

        unvalid_values = F.when(F.col(field.column).isNull(), field.column)
        if isNumricCol:
            unvalid_values = F.when(condition=~valid_condition, value=field.column)

        valid_values = F.when(condition=valid_condition, value=all_values)
        return [
            F.count(valid_values).alias(create_alias("num_samples", field)),
            F.count(unvalid_values).alias(create_alias("num_missing_values", field)),
        ]

    def _aggregate_categorical_column(self, field: PySparkFieldMetadata) -> List[Column]:
        column = field.column
        valid_values = self._get_column_valid_values(F.col(column))

        return [
            F.count(valid_values).alias(create_alias("num_samples", field)),
            F.approx_count_distinct(valid_values).alias(create_alias("num_unique_values", field)),
            F.count(F.when(F.isnan(column) | F.col(column).isNull(), column)).alias(
                create_alias("num_missing_values", field)
            ),
        ]

    def _aggregate_any_column(self, field: PySparkFieldMetadata) -> List[Column]:
        column = field.column
        all_values = F.col(field.column)
        valid_condition = ~all_values.isNull()

        valid_values = F.when(condition=valid_condition, value=all_values)
        return [
            F.count(valid_values).alias(create_alias("num_samples", field)),
            F.count(F.when(F.col(column).isNull(), column)).alias(
                create_alias("num_missing_values", field)
            ),
        ]


def calculate_training_fields(
    data: DataFrame, fields: List[PySparkFieldMetadata]
) -> Dict[str, List[FieldTrainingData]]:
    """Calculates training aggregations on columns in a pyspark dataframe.

    Args:
        data: The dataframe representing the training data
        fields: A list of fields meta data that will be used to calculate the aggregations.

    Returns:
        the aggregation results grouped by the field groups
    """
    aggregator = _TrainingAggregatorPySpark(data=data, fields=fields)
    return aggregator.aggregate_training_fields()



from typing import Optional

from aporia.core.types.field import FieldType


class PySparkFieldMetadata:
    """Field metadata for pyspark training calculations."""

    def __init__(
        self, name: str, type: FieldType, column: str, group: str, key_at_dict: Optional[str] = None
    ):
        """Initializes the object.

        Args:
            name: Field name
            type: Field type
            column: Column corresponding to the field in a dataframe
            group: Field group (features, predictions, etc.)
            key_at_dict: Key of nested field in dict
        """
        self.name = name
        self.type = type
        self.column = column
        self.group = group
        self.key_at_dict = key_at_dict

    def __str__(self) -> str:
        """Returns a string representation of the object."""
        key_at_dict_txt = None
        if self.key_at_dict is not None:
            key_at_dict_txt = f"| Key At Parent Dict Field: {self.key_at_dict}"
        return f"Field Name: {self.name} | Field type: {self.type} | Field Col: {self.column} | Field Group: {self.group} {key_at_dict_txt}"  # noqa: B950
